#include <stdio.h>
#include <assert.h>
#define MAXWIDTH 16
#define epsilon (float)1e-5
#define DATA float

void MatrixMulOnHost(DATA* M, DATA* N, DATA* P, int Width) {
    for (int i = 0; i < Width; ++i) {
        for (int j = 0; j < Width; ++j) {
            DATA pvalue = 0;
            for (int k = 0; k < Width; ++k) {
                DATA a = M[i * Width + k];
                DATA b = N[k * Width + j];
                pvalue += a * b;
            }
        P[i * Width + j] = pvalue;
        }
    }
}

__global__ void MatrixMulKernel(DATA* dM, DATA* dN, DATA* dP, int Width) {
    // Pvalue is used to evaluate the element of the matrix
    // assegnato al thread
   DATA Pvalue = 0.0;
   for (int k = 0; k < Width; ++k) {
      DATA Melement = dM[];
      DATA Nelement = dN[];
      Pvalue += Melement * Nelement;
   }
   dP[] = Pvalue;
}


void MatrixMulOnDevice(DATA* M, DATA* N, DATA* P, int Width) {

    int size = Width * Width * sizeof(DATA);
    DATA *dM, *dN, *dP;

  // Alloc and load input matrices M and N on the GPU memory
    cudaMalloc(&dM, size);
    cudaMemcpy(dM, M, size, cudaMemcpyHostToDevice);

    cudaMalloc(&dN, size);
    cudaMemcpy(dN, N, size, cudaMemcpyHostToDevice);

  //Alloc the output matrix P
    cudaMalloc(&dP, size);
  // Setup the execution configuration
  // Use ONE block (grid is equal to 1x1).
  // Setup the block!
    dim3 grid;
    dim3 block;
    grid.x=1;
    grid.y=1;
    block.x=;
    block.y=;

    // run the kernel!
    MatrixMulKernel<<<grid, block>>>(dM, dN, dP, Width);

  // Copy the result P from the GPU memory
    cudaMemcpy(P, dP, size, cudaMemcpyDeviceToHost);

  // Free the memory
    cudaFree(dM);
    cudaFree(dN);
    cudaFree(dP);

}


// main

int main(int argc, char** argv) {

  int Width;

  DATA *M, *N, *hP, *gP;

  if(argc<2) {
     fprintf(stderr,"Usage: %s Width\n",argv[0]);
     exit(1);
  }

  Width=atoi(argv[1]);

  if(Width<1 ||Width>MAXWIDTH) {
     fprintf(stderr,"Error Width=%d, must be > 0 and <%d\n",Width,MAXWIDTH);
     exit(1);

  }

  M=(DATA *)malloc(Width*Width*sizeof(DATA));
  N=(DATA *)malloc(Width*Width*sizeof(DATA));
  hP=(DATA *)malloc(Width*Width*sizeof(DATA));
  gP=(DATA *)malloc(Width*Width*sizeof(DATA));

  if(M==NULL) {
    fprintf(stderr,"Could not get memory for M\n");
    exit(1);
  }
  if(N==NULL) {
    fprintf(stderr,"Could not get memory for N\n");
    exit(1);
  }
  if(hP==NULL) {
    fprintf(stderr,"Could not get memory for hP\n");
    exit(1);
  }
  if(gP==NULL) {
    fprintf(stderr,"Could not get memory for gP\n");
    exit(1);
  }

  memset(gP,0,Width*Width*sizeof(DATA));
  memset(hP,0,Width*Width*sizeof(DATA));

  for(int y=0; y<Width; y++){
    for(int x=0; x<Width; x++) {
       M[y*Width+x]=(DATA)(y*Width+x);
       N[y*Width+x]=(DATA)(y*Width+x);
    }
  }

  MatrixMulOnHost(M, N, hP, Width);
  MatrixMulOnDevice(M, N, gP, Width);
#if defined(VERBOSE)
  printf("\n\nInput Matrix");
  for(int y=0; y<Width; y++){
    printf("\n");
    for(int x=0; x<Width; x++) {
       printf("%f ", M[y*Width+x]);
    }
  }


  printf("\n\nplain C");
  for(int y=0; y<Width; y++){
    printf("\n");
    for(int x=0; x<Width; x++) {
       printf("%f ", hP[y*Width+x]);
    }
  }

  printf("\n\nGPU C");
  for(int y=0; y<Width; y++){
    printf("\n");
    for(int x=0; x<Width; x++) {
       printf("%f ", gP[y*Width+x]);
    }
  }
#endif
  int errCnt = 0;
  for(int y=0; y<Width; y++){
   for(int x=0; x<Width; x++) {
      DATA it = hP[y*Width+x];
      if(fabs(it - gP[y*Width+x])> epsilon*it)
       errCnt++;

   }
  }

  if(errCnt==0)
    printf("\nTEST PASSED\n");
  else
    printf("\n\nTEST FAILED: number of errors:  %d\n", errCnt);

}
