#include <stdio.h>
#include <math.h>

#define N  (512)

__global__ void dot( int *a, int *b, int *c ) {
   __shared__ int temp[N];
   temp[threadIdx.x] = a[threadIdx.x] * b[threadIdx.x];
       
       // Thread 0 sums the pairwise products
   if( 0 == threadIdx.x ) {
       int sum = 0;
       for( int i = N-1; i >= 0; i-- ){
            sum += temp[i];
	    } 
       *c = sum;
   }
}

void random_ints(int *p, int n) {
	int i;
	for(i=0; i<n; i++) {
		p[i]=rand();
	}
}

int main( void ) {
    int *a, *b, c, d;             // host copies of a, b, c
    int *dev_a, *dev_b, *dev_c;   // device copies of a, b, c
    int size = N * sizeof( int ); // we need space for N int
    int i;

    // allocate device copies of a, b, c
    cudaMalloc( (void**)&dev_a, size );
    cudaMalloc( (void**)&dev_b, size );
    cudaMalloc( (void**)&dev_c, sizeof(int) );

    a = (int*)malloc( size ); 
    b = (int*)malloc( size );

    random_ints( a, N ); 
    random_ints( b, N );
    // copy inputs to device
   cudaMemcpy( dev_a, a, size, cudaMemcpyHostToDevice );
   cudaMemcpy( dev_b, b, size, cudaMemcpyHostToDevice );

    // launch a dot() kernel with N threads
    dot<<< 1, N >>>( dev_a, dev_b, dev_c );

    // copy device result back to host copy of c
   cudaMemcpy( &c, dev_c,sizeof(int),   cudaMemcpyDeviceToHost );

    for(i=0, d=0; i<N; i++) {
	d+=a[i]*b[i];
    }
    if(d!=c) {
	   printf("error: expected %d, got %d!\n",d, c);
    } else {
	   printf("correct!\n");
    }    

 
    free( a ); free( b ); 
    cudaFree( dev_a );
    cudaFree( dev_b );
    cudaFree( dev_c );
    return 0;
}
